export declare const DEFAULT_VERSION = "realtime-js/2.97.0";
export declare const VSN_1_0_0: string;
export declare const VSN_2_0_0: string;
export declare const DEFAULT_VSN: string;
export declare const VERSION = "2.97.0";
export declare const DEFAULT_TIMEOUT = 10000;
export declare const WS_CLOSE_NORMAL = 1000;
export declare const MAX_PUSH_BUFFER_SIZE = 100;
export declare enum SOCKET_STATES {
    connecting = 0,
    open = 1,
    closing = 2,
    closed = 3
}
export declare enum CHANNEL_STATES {
    closed = "closed",
    errored = "errored",
    joined = "joined",
    joining = "joining",
    leaving = "leaving"
}
export declare enum CHANNEL_EVENTS {
    close = "phx_close",
    error = "phx_error",
    join = "phx_join",
    reply = "phx_reply",
    leave = "phx_leave",
    access_token = "access_token"
}
export declare enum TRANSPORTS {
    websocket = "websocket"
}
export declare enum CONNECTION_STATE {
    Connecting = "connecting",
    Open = "open",
    Closing = "closing",
    Closed = "closed"
}
//# sourceMappingURL=constants.d.ts.map